/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Gnu General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Gnu General Public License for more details.
 *
 * You should have received a copy of the Gnu General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antsim.impl;

import anthill.*;
import anthill.nest.*;
import anthill.factory.*;
import anthill.antsim.*;
import anthill.util.*;

import java.util.*;


/**
 *  Class <CODE>ScenarioImpl</CODE> is a simple implementation of
 *  interface <CODE>Scenario</CODE> in which a predefined set of
 *  resources is inserted in the network and then a certain number
 *  of requests are performed and simulate. A <CODE>ScenarioImpl</CODE>
 *  is composed of a <CODE>Peernet</CODE> and a <CODE>ResourceSet</CODE> 
 *  object. Each resource included in the <CODE>ResourceSet</CODE> is inserted
 *  before starting the simulation in the <CODE>Peernet</CODE>. Then,
 *  the <CODE>ResourceSet</CODE> is used to generate a certain number
 *  of requests to be presented to the nests composing the network.
 *  
 *  
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class ScenarioImpl
implements Scenario, ResponseListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Random number generator */
  protected static Random random = new Random();
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  protected Factory factory;
  
  /** Number of rounds to be performed */
  protected int rounds;

  /** Stats */
  protected Stats stats;

  /** Peernet on which the genetic code will be evaluated */
  protected Peernet net;
  
  /** Version identifier of the peer network; when init params change, the version increase */
  protected int netVersion;
  
  /** Set of resources stored in the Anthill network */
  protected ResourceSet resources;
  
  /** Version identifier of the resource set; when init params change, the version increase */
  protected int rsetVersion;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor / reinitialization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a new ScenarioImpl.
   *
   *  @param factory  the factory object used to generate the peernet
   *    and the resource set contained in a scenario
   *  @param rounds the number of rounds of the run
   */
  public ScenarioImpl(Factory factory, int rounds)
  {
    /* Initializes variables */
    this.factory = factory;
    this.rounds = rounds;
    
    /* Invokes clear*/
    clear();
  }

  
  /**
   *  Restores the object to its initial state. The network is re-initialized
   *  as an empty network.
   */
  public void clear()
  {
    int version; 

    /* 
     * Creates a new peernet iff (i) there is no peernet currently; 
     * (ii) the peernet initialization data have changed.
     * Otherwise, clear existing one.
     */
    version = factory.getVersion(Peernet.class);
    if (net == null || netVersion != version ) {
        netVersion = version;
        net = (Peernet) factory.get(Peernet.class);
    } else {
      net.clear();
    }
    
    /* 
     * Creates a new resource set iff (i) there is no resource set currently; 
     * (ii) the resource set initialization data have changed. 
     * Otherwise, clear existing one.
     */
    version = factory.getVersion(ResourceSet.class);
    if (resources == null || rsetVersion != version) {
      rsetVersion = version;
      resources = (ResourceSet) factory.get(ResourceSet.class);
    } else {
      resources.clear();
    }

    /* Inserts generated resources in nests */
    Nest[] nests = net.getNests();
    
    Resource[] rs = resources.getResources();
    for (int i=0; i < rs.length; i++) {
      int pos = random.nextInt(nests.length);  
      nests[pos].insert(rs[i]);
    }
  
  }
   
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Scenario
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Set the statistics object which will be used throughout
   *  the scenario evaluation.
   */
  public void setStats(Stats stats)
  {
    this.stats = stats;
  }
  

  /**
   *  Perform the evaluation of this scenario object by simulating it.
   *  The simulation is subdivided in rounds. At each iteration,
   *  a new request is generated and assigned to one of the nest
   *  included in the simulated peer network, which generates an ant
   *  for it. Furthermore, for each  nest, an ant included in the 
   *  manager queue is selected and its <CODE>run()</CODE> method
   *  is executed. Statistics about the simulated execution of the
   *  code are collected in the <CODE>Stats</CODE> object specified
   *  through method <CODE>setStats()</CODE>.
   */
  public void evaluate()
  {
    // Obtain information about the network
    Nest[] nests = net.getNests();
    Manager[] managers = new Manager[nests.length];
    for (int i=0; i < nests.length; i++) {
      managers[i] = nests[i].getManager();
    }
    
    // Execute the required rounds 
    for (int i=0; i < rounds; i++) {

      /* Choose random nest and make a request */
      int nest = random.nextInt(nests.length);
      Object request = resources.getRequest();
      nests[nest].request(request, this);
      stats.inc("RequestedResources");
      // Log.println(1000, "Searching: " + resource);
      
      /* Each nest executes the code of an ant in its queue */
      for (int j=0; j < managers.length; j++)
        managers[j].execute();
    }
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ResponseListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Upcall method used to notify this listener that the request 
   *  associated to it has returned the specified set of resources.
   * 
   *  @param resources the resources found.
   */
  public void returnResources(Resource[] resources)
  {
    Log.println(1000, "Found: " + resources[0]);
    stats.inc("FoundResources");
  }

  /**
   *  Upcall method used to notify this listener that the request 
   *  associated to it has finished to return resources.
   */
  public void complete()
  {
  }
  
} // END ScenarioImpl

