/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Gnu General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Gnu General Public License for more details.
 *
 * You should have received a copy of the Gnu General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antsim.impl;

import java.util.*;
import anthill.util.*;
import anthill.factory.*;
import anthill.antsim.*;

/**
 *  The <I>Stats</I> class enables programmers to collect information
 *  about events occurring during a simulation. In particular, this
 *  class may be used to count the occurrence of events. Each event
 *  is identified by a string name chosen by the programmer and is 
 *  associated to a counter. The value of the counter may be incremented
 *  by one by invoking method <I>increment()</I>; the same value
 *  may be read by invoking method <I>get()</I>.
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class StatsImpl
implements Stats, Clearable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Hash table mapping event names and event counters.
   */
  private HashMap map;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Initializes an empty stats object.
   */
  public StatsImpl()
  {
    map = new HashMap();
  }
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Increment by 1 the value of the counter identified by <CODE>name</CODE>.
   *
   */
  public void inc(String name)
  {
    add(name, 1);
  }

  /**
   *  Increment by 1 the value of the counter identified by <CODE>name</CODE>.
   *
   */
  public void add(String name, float addend)
  {
    Wrapper wrapper = (Wrapper) map.get(name);
    if (wrapper == null) {
      wrapper = new Wrapper();
      map.put(name, wrapper);
    }
    wrapper.value += addend;
  }

  /**
   *  Return the value of the counter identified  by <CODE>name</CODE>
   */
  public float get(String name)
  {
    Wrapper wrapper = (Wrapper) map.get(name);
    return (wrapper != null ? wrapper.value : -1);
  }


  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the object to its initial state, setting all fields
   *  to zero.
   */
  public void clear()
  {
    map.clear();
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Wrapper inner class
  ////////////////////////////////////////////////////////////////////////////////////////////

  private class Wrapper
  {
    /** Float value enclosed in this wrapper */
    private float value;
  
    /** Default constructor */
    private Wrapper() 
    { 
      value = 0;
    }
  } // END Wrapper
  
} // END Stats

