/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.factory;

import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

/**
 *  Parser code to read an XML configuration file, using the supplied URL.
 *  The default URL is 'file:$HOME/.anthill/config.xml', which is used if the
 *  non-argument constructor is used. Otherwise, the constructor takes the
 *  URL as an argument.
 *
 *  @author Hein Meling
 *  @version $REVISION$
 */
public final class ConfigParser {

  /////////////////////////////////////////////////////////////////////////////
  // Constants - Error message and XML tags
  /////////////////////////////////////////////////////////////////////////////

  private static final String cfgRootTagName        = "SystemConfiguration";
  private static final String cfgVersionAttrib      = "version";
  private static final String cfgVersionValue       = "1.0";

  private static final String cfgCreateTagName      = "Create";
  private static final String cfgCreateIfaceAttrib  = "interface";
  private static final String cfgCreateClassAttrib  = "class";

  private static final String cfgArgTagName         = "Argument";
  private static final String cfgArgNameAttrib      = "name";
  private static final String cfgArgTypeAttrib      = "type";
  private static final String cfgArgValueAttrib     = "value";

  /////////////////////////////////////////////////////////////////////////////
  // Static block - Default URL for the configuration file
  /////////////////////////////////////////////////////////////////////////////

  private static final String configURL;

  /*  Initialize the default URL for the configuration file.
   *  Used if no URL is supplied to the constructor. */
  static {
    String url = "file:" + System.getProperty("user.home");
    String s = java.io.File.separator;
    configURL = url.concat(s + ".anthill" + s + "config.xml");
  }

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  /*  Factory object used to invoke this parser object. */
  private Factory factory;

  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a parser that parses the file given by the default URL.
   *
   * @param factory    The factory object invoking the parser. Used to update
   *                   the factory's hash map of interface to <class, arg>
   */
  public ConfigParser(Factory factory)
    throws ConfigFormatException, java.io.IOException
  {
    this(configURL, factory);
  }

  /**
   * Constructs a parser that parses the file obtained from the specified URL.
   *
   * @param factory    The factory object invoking the parser. Used to update
   *                   the factory's hash map of interface to <class, arg>
   * @param urlConfig  The URL to the xml configuration file containing
   *                   the description of the distributed system.
   */
  public ConfigParser(String urlConfig, Factory factory)
    throws ConfigFormatException, java.io.IOException
  {
    this.factory = factory;
    Document doc = null;
    try {
      DocumentBuilderFactory docBF = DocumentBuilderFactory.newInstance();
      docBF.setValidating(true);
      DocumentBuilder db = docBF.newDocumentBuilder();
      doc = db.parse(urlConfig);
    } catch (Exception exception) {
      exception.printStackTrace();
      throw new java.io.IOException("Unable to set up parsing environment");
    }
    Element elm = doc.getDocumentElement();
    checkCfgVersion(elm);
    NodeList nl = elm.getChildNodes();
    for (int i=1; i<=nl.getLength(); i++) {
      Node node = nl.item(i);
      if (node == null || node.getNodeName().startsWith("#")) {
        // Node is not an element, move on to the next node.
        continue;
      }
      elm = (Element)node;
      // This will place all mappings directly into the factory hash table.
      getCreateTag(elm, factory);
    }
  }


  /////////////////////////////////////////////////////////////////////////////
  // Private methods for parsing the XML tree
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Check that the configuration file has the expected
   *  version number and root tag.
   */
  private void checkCfgVersion(Element elm)
    throws ConfigFormatException
  {
    String cfgRoot = elm.getTagName();
    String cfgVersion = elm.getAttribute(cfgVersionAttrib);
    if (cfgRoot.equalsIgnoreCase(cfgRootTagName)) {
      if (cfgVersion.equals(cfgVersionValue)) {
        // All is OK, continue parsing.
        return;
      } else {
        throw new ConfigFormatException("Unexpected version ("
          + cfgVersion + "). Expected version " + cfgVersionValue);
      }
    } else {
      throw new ConfigFormatException("Unexpected root tag ("
        + cfgRoot + "). Expected " + cfgRootTagName);
    }
  }


  /**
   *  Get the content of a <Create> tag, including its <Argument> tag.
   *  
   */
  private void getCreateTag(Element elm, Factory factory)
    throws ConfigFormatException
  {
    Object arg;
    if (elm != null && elm.getTagName().equalsIgnoreCase(cfgCreateTagName)) {
      String iface = elm.getAttribute(cfgCreateIfaceAttrib);
      String clazz = elm.getAttribute(cfgCreateClassAttrib);
      if (iface.length() == 0) {
        throw new ConfigFormatException(
          "Required 'interface' attribute not specified.");
      }
      if (clazz.length() == 0) {
        throw new ConfigFormatException(
          "Required 'class' attribute not specified.");
      }
      // System.out.println("iface: " + iface + ", class: " + clazz);
      NodeList argList = elm.getChildNodes();

      // Allocate space in the ArrayList to store the complete argument list
      // ALB comment, 29-05-01: ArrayList needed because some of the items
      // in argList could be not elements.
      ArrayList aList = new ArrayList(argList.getLength());

      for (int i=0; i<argList.getLength(); i++) {
        Node node = argList.item(i);
        if (node == null || node.getNodeName().startsWith("#")) {
          // Node is not an element, move on to the next node.
          continue;
        }
        Element elem = (Element)node;
        arg = getArgumentTag(elem);
        if (arg == null)
          arg = factory;
        aList.add(arg);
      }
      // Convert the argument list to an array of Objects
      Object[] args = (Object[]) aList.toArray(new Object[aList.size()]);
      // Put the mapping into the internal factory ifaceMap table.
      factory.put(iface, clazz, args);
    } else {
      throw new ConfigFormatException("Expected <Create> tag.");
    }
  }


  /**
   *  Get the content of a single <Argument> tag.
   *
   *  @param elm  An element in the DOM tree.
   *  @return     Argument converted to its value
   */
  private Object getArgumentTag(Element elm)
    throws ConfigFormatException
  {
    if (elm != null && elm.getTagName().equalsIgnoreCase(cfgArgTagName)) {
      String name  = elm.getAttribute(cfgArgNameAttrib);
      String type  = elm.getAttribute(cfgArgTypeAttrib);
      String value = elm.getAttribute(cfgArgValueAttrib);
      if (name.length() == 0) {
        throw new ConfigFormatException(
          "Required 'name' attribute not specified.");
      }
      if (type.length() == 0) {
        throw new ConfigFormatException(
          "Required 'type' attribute not specified.");
      }
      /* ALB Patch, 29-05-01 
      if (value.length() == 0) {
        throw new ConfigFormatException(
          "Required 'value' attribute not specified.");
      }
      */
      // System.out.println("name: " + name + ", type: " + type + ", value: " + value);
      /* Note that the 'name' parameter is not passed along here
       * since we don't need it. It is only used for user interpretation
       * of the XML configuration file. */
      return convertArg(type, value);
    } else {
      throw new ConfigFormatException("Expected <Argument> tag.");
    }
  }


  private Object convertArg(String prim_type, String value)
    throws ConfigFormatException
  {
    Object arg;
    if (prim_type.equals("byte")) {
      arg = (Object) new Byte(value);
    } else if (prim_type.equals("char")) {
      // Note, we assume the first char of a string value is to be used
      // Maybe we could provide a check on the length of the string
      // and throw an exception if it is longer than 1 character.
      arg = (Object) new Character(value.charAt(0));
    } else if (prim_type.equals("double")) {
      arg = (Object) new Double(value);
    } else if (prim_type.equals("float")) {
      arg = (Object) new Float(value);
    } else if (prim_type.equals("int")) {
      arg = (Object) new Integer(value);
    } else if (prim_type.equals("long")) {
      arg = (Object) new Long(value);
    } else if (prim_type.equals("short")) {
      arg = (Object) new Short(value);
    } else if (prim_type.equals("boolean")) {
      arg = (Object) new Boolean(value);
    } else if (prim_type.equals("String")) {
      arg = (Object) new String(value);
    } else if (prim_type.equals("Factory")) {
      arg = null;
    } else {
      throw new ConfigFormatException("Only primitive types can be used on"
        + " constructors using the generic factory.");
    }
    return arg;
  }

} // END of class ConfigParser
