/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.factory;

import anthill.*;

import java.util.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

/**
 *  A generic factory for creating object instances of a class
 *  matching a particular interface and possibly taking arguments.
 *  The mapping between an implementing class and an interface,
 *  and also the arguments to be used to instantiate the class
 *  are all read from a configuration file.
 *
 *  @author Hein Meling
 *  @version $REVISION$
 */
public class Factory
{
  /////////////////////////////////////////////////////////////////////////////
  // Constants - Error message
  /////////////////////////////////////////////////////////////////////////////

  private static final String errorMsg = "Error in configuration file: ";

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*  Map of interfaces mapping to class, args pairs. */
  private HashMap ifaceMap = new HashMap();

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Create a parser object and pass this factory object allowing the
   *  parser to place the parsed elements directly into the HashMap
   *  using the factory API method <CODE>put</CODE>.  The parser will
   *  use the provide URL for the configuration file location.
   *
   *  @param configURL  A URL pointing to the configuration file to parse.
   *
   */
  public Factory(String configURL)
    throws ConfigFormatException, java.io.IOException
  {
    ConfigParser cp = new ConfigParser(configURL, this);
  }

  /**
   *  Create a parser object and pass this factory object allowing the
   *  parser to place the parsed elements directly into the HashMap
   *  using the factory API method <CODE>put</CODE>.  The parser will
   *  use the default configuration file location.
   *
   */
  public Factory()
    throws ConfigFormatException, java.io.IOException
  {
    ConfigParser cp = new ConfigParser(this);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Factory interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The <CODE>get</CODE> method is used to obtain an instantiation of
   *  a class implementing the requested interface.
   *
   *  @param cl The interface used as mapping key.
   *  @return   An object instance implementing the specified interface.
   *
   *  @throws ConfigFormatException  Thrown if specified classes and interfaces
   *     cannot be found.
   *  @throws ConstructorException Thrown if the constructor reflectively invoked
   *    by get raises an exception
   */
  public Object get(Class cl)
    throws ConfigFormatException, ConstructorException
  {
    return get(cl.getName());
  }

  /**
   *  The <CODE>get</CODE> method is used to obtain an instantiation of
   *  a class implementing the requested interface.
   *
   *  @param iface  The interface name used as mapping key.
   *  @return       An object instance implementing the specified interface.
   *
   *  @throws ConfigFormatException  Thrown if specified classes and interfaces
   *     cannot be found.
   *  @throws ConstructorException Thrown if the constructor reflectively invoked
   *    by get raises an exception
   */
  public Object get(String iface)
    throws ConfigFormatException, ConstructorException
  {
    // Obtain the ClassData for the specified interface
    ClassData cdata = (ClassData) ifaceMap.get(iface);
    if (cdata == null) {
      throw new ConfigFormatException(errorMsg + "interface not defined: " + iface);
    }

    try {
      // Get a Class object representing the specified class.
      Class clasz = Class.forName(cdata.clazz);
      // Obtain a constructor for the set of arguments used in the constructor.
      Constructor c = clasz.getConstructor(cdata.types);
      // Invoke constructor with the specifed args
      // Note that the newInstance method will unwrap the primitive types.
      // Thus, it doesn't matter that they are not
      // "int" as long as they are "Integer".
      return (Object) c.newInstance(cdata.args);
    } catch (ClassNotFoundException exception) {
      throw new ConfigFormatException(errorMsg
        + "No matching class " + cdata.clazz + " was found.");
    } catch (NoSuchMethodException exception) {
      throw new ConfigFormatException(errorMsg
        + "Found no matching constructor for class " + cdata.clazz);
    } catch (InstantiationException exception) {
      throw new ConfigFormatException(errorMsg
        + "The specified class " + cdata.clazz
        + " is an interface or abstract class.");
    } catch (IllegalAccessException exception) {
      throw new ConfigFormatException(errorMsg
        + "The specified class " + cdata.clazz + " is not accessible.");
    } catch (InvocationTargetException exception) {
      Throwable e = exception.getTargetException();
      while (e instanceof ConstructorException) {
        e = ((ConstructorException) e).getTargetException();
      }
      throw new ConstructorException(e);
    }
  }


  /**
   *  Check the version of the class, argument pair stored in the
   *  factory database under the key <CODE>iface</CODE>.
   *
   *  @param cl The interface to check the current version of.
   *
   */
  public int getVersion(Class cl) 
  {
    return getVersion(cl.getName());
  }
  
  /**
   *  Check the version of the class, argument pair stored in the
   *  factory database under the key <CODE>iface</CODE>.
   *
   *  @param iface  The interface to check the current version of.
   *
   */
  public int getVersion(String iface) {
    if (ifaceMap.containsKey(iface)) {
      ClassData cdata = (ClassData) ifaceMap.get(iface);
      return cdata.version;
    } else {
      return -1;
    }
  }


  /**
   *  Method used to insert or update an interface mapping in the
   *  factory database.
   *
   *  @param iface  The interface name used as mapping key.
   *  @param clazz  The class name implementing the above interface.
   *  @param args   Array of arguments used for instantiating the above class.
   *
   */
  public void put(String iface, String clazz, Object[] args) {
    // ALB FIX, 29-05-01: moved this code from get to put; in this way,
    // classes are computed only once 
    // Loop to evaluate the type of each argument in the argument list
    // This is needed for the constructor of the object represented by
    // the class stored in the ifaceMap.
    Class[] types = new Class[args.length];

    for (int i=0; i < args.length; i++) {
      String type = args[i].getClass().getName();
      if (type.equals("java.lang.Byte")) {
        types[i] = byte.class;
      } else if (type.equals("java.lang.Character")) {
        types[i] = char.class;
      } else if (type.equals("java.lang.Double")) {
        types[i] = double.class;
      } else if (type.equals("java.lang.Float")) {
        types[i] = float.class;
      } else if (type.equals("java.lang.Integer")) {
        types[i] = int.class;
      } else if (type.equals("java.lang.Long")) {
        types[i] = long.class;
      } else if (type.equals("java.lang.Short")) {
        types[i] = short.class;
      } else if (type.equals("java.lang.Boolean")) {
        types[i] = boolean.class;
      } else if (type.equals("java.lang.String")) {
        types[i] = String.class;
      } else if (type.equals(Factory.class.getName())) {
        types[i] = Factory.class;
      } else {
        System.err.println("Internal error, type unknown: " + type);
      }
    } // END of for loop
    if (ifaceMap.containsKey(iface)) {
      // Updating an existing interface entry with new class and new parameters
      ClassData cdata = (ClassData) ifaceMap.get(iface);
      cdata.update(clazz, types, args);
    } else {
      ifaceMap.put(iface, new ClassData(clazz, types, args));
    }
  }


  /////////////////////////////////////////////////////////////////////////////
  // Main method - used only for testing
  /////////////////////////////////////////////////////////////////////////////

  /*
  public static void main(String[] args)
    throws Exception
  {
    Factory f = null;
    if (args.length >= 1) {
      f = new Factory(args[0]);
    } else {
      f = new Factory();
    }
    Gate g = (Gate) f.get("anthill.Gate");
    System.out.println("Gate: " + g);
    AnthillId aid = g.getId();
    System.out.println("Aid: " + aid);
    f.put("anthill.Gate", "anthill.antnet.UdpGate", new Object[] {new Integer(23)});
    System.out.println("anthill.Gate version: " +f.getVersion("anthill.Gate"));
  }
  */
}
