/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Gnu General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Gnu General Public License for more details.
 *
 * You should have received a copy of the Gnu General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nest;

import anthill.*;
import anthill.util.*;

/**
 *  The <CODE>Storage</CODE> interface represents the resource storage
 *  component of a nest. Resources maintained by a <CODE>Storage</CODE> 
 *  object may be retrieved by invoking the <CODE>getResource()</CODE> 
 *  method. Resources may be added to a storage by invoking methods
 *  <CODE>addAntResource()</CODE> and <CODE>addClientResource()</CODE>.
 *  The former method is invoked by ants through their associated
 *  <CODE>AntView</CODE> object, while the latter is invoked by nest
 *  users. 
 *  <p>
 *  Depending on the implementation, different storage policies
 *  may be adopted for resources inserted by ants and users.
 *  For example, ants resources may be handled with a LRU policy,
 *  while user resources may be stored permanently.
 *  <p>
 *  <CODE>Storage</CODE> is one of the components of a nest; the
 *  other are <CODE>Manager</CODE>), <CODE>RequestSet</CODE>, 
 *  <CODE>Gate</CODE> and <CODE>Trail</CODE>.
 *  <p>
 *  <CODE>Storage</CODE> methods are invoked by <CODE>AntView</CODE>
 *  objects to implement methods <CODE>addResource()</CODE>
 *  and <CODE>getResources()</CODE>. The exact behavior of
 *  these methods depends on the implementation of <CODE>Storage</CODE>.

 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public interface Storage
extends Clearable
{

  /**
   *  Returns the set of resources stored locally which satisfy
   *  the request. If no resource can be found, a <CODE>null</CODE> 
   *  value is returned.
   *
   *  @param request the request to be satisfied
   */
  public Resource[] getResources(Object request);

  /**
   *  Adds a resource to the storage. This method is invoked 
   *  from a nest on behalf of an ant. 
   *
   *  @param resource the resource to be added
   */
  public void addAntResource(Resource resource);
  
  /**
   *  Add a resource to the storage. This method is invoked 
   *  from a nest on behalf of a nest user. 
   *
   *  @param resource the resource to be added
   */
  public void addClientResource(Resource resource);
  
} // END Storage
