/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Gnu General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Gnu General Public License for more details.
 *
 * You should have received a copy of the Gnu General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nestor;

import anthill.*;
import anthill.nest.*;
import anthill.factory.*;
import anthill.genie.*;

import java.util.*;

/**
 *  A Nest class collects the components of a Nest site: (i) a AntView
 *  object, which is the view of a Nest for an Ant; (ii) a Gate object,
 *  which manage the communication between nests; (iii) an Manager,
 *  i.e. the object which manage the activities performed by different agents.
 *  <p>
 *  Implementation note: this implementation is supposed to be used as
 *  simple implementation of <CODE>Nest</CODE> in the off-line
 *  environment. Here, we assume that the Manager object used is single-threaded;
 *  in this way, we avoid the costs of synchronized methods.
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class LocalNest
implements Nest, AntListener
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Initialization fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  private Factory factory;
  
  /** Time-to-live (FIXME fixed, should be configurable) */
  int ttl;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Ant factory used to generate ants for request */
  private AntFactory antf;
  
  /** Reference to the resource database */
  private Storage storage;
  
  /** Communication handler to be used to communicate with other nests */
  private Gate gate;
  
  /** FIXME */
  private RequestSet requests;

  /** Thread manager needed to be used to manage concurrent agents */
  private Manager manager;

  /** Back trail information */
  private Trail trail;
  
  /** Neighbours hash table, with NestId as key */
  private HashMap map;

  /** Neighbours array */
  private Neighbour[] neighbours;

  /** Pheromone object managed by ants */
  private Object pheromone;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a new nest, by putting together a Storage
   *  (a resource database), a Gate (i.e. the basic API for 
   *  communicating with other nests) and an Manager
   *  (i.e., a thread handler managing the agent execution).
   */
  public LocalNest(Factory factory, int ttl)
  {
    /* Update factories */
    this.factory = factory;
    antf = (AntFactory) factory.get(AntFactory.class);

    /* Obtain implementations of basic interfaces */
    storage = (Storage) factory.get(Storage.class);
    trail = (Trail) factory.get(Trail.class); 
    gate = (Gate) factory.get(Gate.class);
    requests = (RequestSet) factory.get(RequestSet.class);
    manager = (Manager) factory.get(Manager.class);

    /* Init other variables */
    map = new HashMap();
    gate.setAntListener(this);
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from AntListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Receive an agent from outside and deliver it to the
   *  to the activity manager.
   */
  public void deliver(AntDescriptor desc)
  {
    AntView view = new LocalAntView(this, desc); 
    manager.add(view);
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Nest
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the identifier of this Nest
   */
  public NestId getId()
  {
    return gate.getId();
  }
  
  /**
   *  Return the Manager associated to this Nest
   */
  public Manager getManager()
  {
    return manager;
  }

  /**
   *  Accepts a request from the nest user. Possible responses will be
   *  returned to the specified listener. One or more ants are generated
   *  and assigned to the request.
   *  Accept a resource request originated by an anthill client.
   *  The resource to be searched is identified by name. The method
   *  returns immediately. When the resource is found, a Resource
   *  object will be returned to the specified resource listener.
   *  <p>
   *  Implementation note: If the resource can be found locally, 
   *  the listener is immediately contacted. Otherwise, a new 
   *  <CODE>Ant</CODE> is generated and is delivered to the ant 
   *  manager.
   *
   *  @param request the request to be performed
   *  @param listener The listener object which will be notified
   *    when the request is completed.
   */
  public void request(Object request, ResponseListener listener)
  {
    /* Creates a set of ants and insert them in the local nest */
    Ant[] ants = antf.getAnts(request);
    for (int i=0; i < ants.length; i++) {
      AntId aid = new LocalAid();
      AntDescriptor desc = new AntDescriptor(gate.getId(), aid, ants[i], ttl);
      AntView view = new LocalAntView(this, desc);
      manager.add(view);
      requests.add(aid, listener);
    } 
  }
  
  /**
   *  Insert a new resource in this anthill. Delagates
   *  the insertion to the storage.
   * 
   *  @param resource the resource to be inserted.
   */
  public void insert(Resource resource)
  {
    storage.addAntResource(resource);
  }

  
  /**
   *  Add the identifier of a new neighbour anthill to the set of 
   *  neighbours known to this <CODE>Nest</CODE>, on behalf of 
   *  an Nest client. 
   *
   *  @param aid neighbour identifier
   */
  public void addClientNeighbour(NestId aid)
  {
    map.put(aid, new Neighbour(aid));
    neighbours = null;
  }
  
  
  /**
   *  Add the identifier of a new neighbour anthill to the set of 
   *  neighbours known to this <CODE>Nest</CODE>, on behalf of 
   *  an ant. 
   *
   *  @param aid neighbour identifier
   */
  public void addAntNeighbour(NestId aid)
  {
    map.put(aid, new Neighbour(aid));
    neighbours = null;
  }

  
  /**
   * Remove a neighbour identifier from the set of neighbour
   * @param aid neighbour identifier
   */
  public void removeNeighbour(NestId aid)
  {
    map.remove(aid);
    neighbours = null;
  }
  
  /**
   * Return the neighbours known to this Nest
   */
  public Neighbour[] getNeighbours()
  {
    if (neighbours == null) {
      neighbours = new Neighbour[map.size()];
      Iterator iterator = map.values().iterator();
      for (int i=0; i < neighbours.length; i++) 
        neighbours[i] = (Neighbour) iterator.next();
    }
    return neighbours;
  }
  
  /**
   *  Return the pheromone associated to the link to this
   *  neighbour. The class of the pheromone is left unspecified.
   */
  public Object getPheromone(Class cl)
  {
    // FIXME Add a per-class table
    return pheromone;
  }

  
  /**
   *  Return the pheromone associated to the link to this
   *  neighbour. The class of the pheromone is left unspecified.
   */
  public void setPheromone(Object pheromone, Class cl)
  {
    // FIXME Add a per-class table
    this.pheromone = pheromone;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the <CODE>Gate</CODE> object used by this <CODE>Nest</CODE>
   */
  Gate getGate()
  {
    return gate;
  }
    
  /**
   *  Return the <CODE>Storage</CODE> object used by this <CODE>Nest</CODE>
   */
  Storage getStorage()
  {
    return storage;
  }
    
  /**
   *  Return the <CODE>RequestSet</CODE> object used by this <CODE>Nest</CODE>
   */
  RequestSet getRequestSet()
  {
    return requests;
  }
  
  /**
   *  Return the <CODE>RequestSet</CODE> object used by this <CODE>Nest</CODE>
   */
  Trail getPreviousNest(boolean create)
  {
    if (trail == null && create) {
      trail = (Trail) factory.get(Trail.class);
    }
    return trail;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the object to its initial state.
   */
  public void clear()
  {
    storage.clear();
    gate.clear();
    manager.clear();
    requests.clear();
    neighbours = new Neighbour[0];
    map.clear();
    gate.setAntListener(this);
  }
  
} // END LocalNest
