/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Gnu General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Gnu General Public License for more details.
 *
 * You should have received a copy of the Gnu General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nestor;

import anthill.*;
import anthill.nest.*;
import anthill.factory.*;

  import java.util.*;

/**
 *  <CODE>LocalStorage</CODE> is a simple implementation of the <CODE>Storage</CODE> 
 *  interface, aimed at serving as resource repository for nests during the off-line
 *  simulation. Each <CODE>LocalStorage</CODE> is characterized by its capacity, i.e. 
 *  by the  number of documents that can be simultaneously maintained in it.
 *  Due to the fact that the capacity is limited, resources are stored
 *  following a last-recently-used (LRU) policy.
 *  <p>
 *  This data structure is designed for single-threaded off-line simulation,
 *  i.e. its methods are not synchronized.
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class LocalStorage
implements Storage
{
  
 
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Current size (i.e., number of resources currently stored in this storage) */
  protected int size; 

  /** Current capacity (i.e., number of resources) of this storage */
  protected int capacity; 
  
  /** LRU queue */
  protected LinkedList list;
  
  /** Hash map for obtaining resources */
  protected HashMap map;
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Builds a new empty storage, with the given capacity.
   *  
   *  @param capacity the number of resources which can be stored in this object
   */
  public LocalStorage(int capacity)
  {
    this.capacity = capacity;
    list = new LinkedList();
    map = new HashMap();
    
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Storage
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returns the <CODE>Resource</CODE> objects which satisfy the specified
   *  request. This implementations assumes that a request for a documents
   *  is constituted by a string containing the object itself. If the resource
   *  is present, it is moved to the first position of the LRU queue.
   *
   *  @param request 
   *    the request to be satisfied 
   *  @return 
   *    the found resources, or null if no resource may be found
   */
  public Resource[] getResources(Object request)
  {
    Resource resource = (Resource) map.get(request);
    if (resource != null) {
      list.remove(request);
      list.addFirst(request);
      return new Resource[] { resource };
    } else {
      return null;
    }
  }

  /**
   *  Add a resource to the storage. This method is invoked 
   *  from a nest on behalf of an ant. If the resource is 
   *  already present in the storage, it is moved to the first
   *  position of the LRU queue. If the resource is not present, 
   *  it is stored in the storage and it is placed in the first 
   *  position of the LRU queue. In both cases, the storage 
   *  verifies that the number of resources in the system does 
   *  not exceed  its current capacity; if so, resources in the 
   *  last position of the LRU queue are removed from the storage.
   *
   *  @param resource the resource to be added
   */
  public void addAntResource(Resource resource)
  {
    // Retrieve the name of the resource
    String name = resource.getName();
    
    // If the resource is present, remove it from its current LRU position
    if (list.remove(name))
      size--;
    
    // Add the resource to the head of the LRU list
    list.addFirst(name);
    map.put(name, resource);
    size++;

    // Remove resources exceeding capacity
    while (size >= capacity) {
      map.remove(list.removeLast());
      size--;
    }
  }
  
  /**
   *  Add a resource to the storage. This method is invoked 
   *  from a nest on behalf of an user. In this simple 
   *  implementation, the semantics of an insertion requested
   *  by an user is equal to the semantics of an insertion
   *  requested by an ant.
   *
   *  @param resource the resource to be added
   */
  public void addClientResource(Resource resource)
  {
    addAntResource(resource);
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Restores the object to its initial state.
   */
  public void clear()
  {
    size = 0;
    map.clear();
    list.clear();
  }

} // END LocalStorage

